//
// Created by lenfrex on 2023/5/7.
//

#include "Grammar.h"

using namespace std;

void Grammar::generateFirstSet() {
    // 处理每一个非终止符
    for (char nonTerminal: nonTerminals) {
        calcFirstSetForNonTerminal(nonTerminal);
    }
}

void Grammar::calcFirstSetForNonTerminal(char nonTerminal) {
    if (!productionsMap.contains(nonTerminal)) {
        return;
    }

    const set<string> &productions = productionsMap.at(nonTerminal);
    // 处理非终止符的每个产生式
    for (const auto &production: productions) {
        for (char currentChar: production) {
            // 如果这个字符是终止符那就直接加到终止符集合里面，并退出逐字符解析
            if (terminals.contains(currentChar)) {
                firstSet[nonTerminal].insert(currentChar);
                break;
            }

            // 该非终止符是否已经有first集合了，有的话就全部加到里面去
            // 否则递归处理，先计算这个非终止符的first
            ProcessNonTerminal:
            if (firstSet.contains(currentChar)) {
                firstSet[nonTerminal].insert(firstSet[currentChar].begin(), firstSet[currentChar].end());
            } else {
                // 如果左递归了，就先处理其他的表达式，但是这个时候其实已经不算是LL1文法了
                if (currentChar == nonTerminal) {
                    break;
                }
                calcFirstSetForNonTerminal(currentChar);
                // 递归完还要回来继续处理之前待处理的非终止符，如S->A..., A->a, 递归处理完A后还要返回继续处理S
                goto ProcessNonTerminal;
            }

            // 如果这个非终止符不会推出空字符，就直接退出解析
            // 能推出空字符就继续
            if (!canDeducedEmpty(productionsMap.at(currentChar))) {
                break;
            }
        }
    }
}